@@ -2,7 +2,7 @@ import csv
 import itertools
 import xlsxwriter
 import xl2dict
-import xlrd
+import xlrd3 as xlrd
 import errno
 import os
 import logging
@@ -10,10 +10,40 @@ import faker
 from random import sample, randint
 import baangt.base.GlobalConstants as GC
 import re
+from openpyxl import load_workbook
 
 logger = logging.getLogger("pyC")
 
 
+class Writer:
+    """
+    This class is made to update existing excel file.
+    First it will open the file in python and then we can do multiple writes and once everything is update we can use
+    save method in order to save the updated excel file. Hence, this class is very useful is saving time while updating
+    excel files.
+    """
+    def __init__(self, path):
+        self.path = path
+        self.workbook = load_workbook(self.path)
+
+    def write(self, row, data, sht):
+        # Update the values using row and col number.
+        # Note :- We are using openpyxl so row & column index will start from 1 instead of 0
+        column = 0
+        sheet = self.workbook[sht]
+        headers = next(sheet.rows)
+        for header in headers:  # checks if usecount header is present in sheet
+            if "usecount" in str(header.value).lower():
+                column = headers.index(header) + 1
+        if column:
+            sheet.cell(row, column).value = data
+
+
+    def save(self):
+        # Call this method to save the file once every updates are written
+        self.workbook.save(self.path)
+        self.workbook.close()
+
 class TestDataGenerator:
     """
     TestDataGenerator Class is to used to create a TestData file from raw excel file containing all possible values.
@@ -39,9 +69,13 @@ class TestDataGenerator:
             raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self.path)
         self.sheet_dict, self.raw_data_json = self.__read_excel(self.path, self.sheet_name)
         self.remove_header = []
+        self.usecount_dict = {}  # used to maintain usecount limit record and verify if that non of the data cross limit
         self.processed_datas = self.__process_data(self.raw_data_json)
         self.headers = [x for x in list(self.processed_datas[0].keys()) if x not in self.remove_header]
+        self.headers = [x for x in self.headers if 'usecount' not in x.lower()]
+        self.writer = Writer(self.path)  # Writer class object to save file only in end, which will save time
         self.final_data = self.__generateFinalData(self.processed_datas)
+        self.writer.save()  # saving source input file once everything is done
 
     def write(self, OutputFormat=GC.TESTDATAGENERATOR_OUTPUT_FORMAT, batch_size=0, outputfile=None):
         """
@@ -203,6 +237,7 @@ class TestDataGenerator:
         :return: None
         """
         for data in data_list:
+            success = True  # if usecount data is present then this is used to keep track of it and not to add whole row in final
             data = list(data)
             done = {}
             for ind in dictionary:
@@ -218,9 +253,28 @@ class TestDataGenerator:
                                 sorted_data = [x for x in dictionary[ind] if x[header] == match]
                                 break
                         if not sorted_data:
-                            sorted_data = dictionary[ind]
-                        data_to_insert = sorted_data[randint(0, len(sorted_data) - 1)]
+                            sorted_data = list(dictionary[ind])
+                        remove_data = []  # Used to remove data with reached limit of usecount
+                        for dtt in sorted_data:  # this loop will check if data has reached the usecount limit and remove
+                            if self.usecount_dict[repr(dtt)]['limit'] == 0:
+                                continue
+                            if not self.usecount_dict[repr(dtt)]['use'] < self.usecount_dict[repr(dtt)]['limit']:
+                                remove_data.append(dtt)
+                        for dtt in remove_data:  # removing data from main data list
+                            logger.debug(f"UseCount limit of {dtt} is exceeded : {str(self.usecount_dict[repr(dtt)]['limit'])}")
+                            sorted_data.remove(dtt)
+                        if len(sorted_data) == 0:  # if the current loop has reached usecount the we need not to add whole row in final output
+                            success = False
+                            break
+                        elif len(sorted_data) == 1:
+                            data_to_insert = sorted_data[0]
+                        else:
+                            data_to_insert = sorted_data[randint(0, len(sorted_data) - 1)]
+                        self.usecount_dict[repr(data_to_insert)]['use'] += 1
                         for keys in data_to_insert:
+                            if "usecount" in keys.lower():  # removing usecount header from headers in final output
+                                self.update_usecount_in_source(data_to_insert)
+                                continue
                             if keys not in self.headers:
                                 self.headers.append(keys)
                             if keys not in done:
@@ -229,7 +283,8 @@ class TestDataGenerator:
                     else:
                         data_to_insert = dictionary[ind][randint(0, len(dictionary[ind]) - 1)]
                         data.insert(ind, data_to_insert)
-            final_list.append(data)
+            if success:
+                final_list.append(data)
 
     def __prefix_data_processing(self, dic, key, dictionary: dict):
         """
@@ -365,6 +420,32 @@ class TestDataGenerator:
             processed_datas = self.__processRrd(first_value, second_value,evaluated_dict)
             processed_datas = data_type(processed_datas)
 
+        elif prefix == "Rre":
+            file_name = raw_data[1:-1].split(',')[0].strip()
+            sheet_dict, _ = self.__read_excel(file_name)
+            first_value = raw_data[1:-1].split(',')[1].strip()
+            second_value = raw_data[1:-1].split(',')[2].strip()
+            if second_value[0] == "[":
+                second_value = ','.join(raw_data[1:-1].split(',')[2:]).strip()
+                second_value = second_value[:second_value.index(']')+1]
+                third_value = [x.strip() for x in ']'.join(raw_data[1:-1].split(']')[1:]).split(',')[1:]]
+            else:
+                third_value = [x.strip() for x in raw_data[1:-1].split(',')[3:]]
+            evaluated_list = ']],'.join(','.join(third_value)[1:-1].strip().split('],')).split('],')
+            if evaluated_list[0] == "":
+                evaluated_dict = {}
+            else:
+                evaluated_dict = {
+                    splited_data.split(':')[0]: self.__splitList(splited_data.split(':')[1])  for splited_data in evaluated_list
+                }
+            if second_value[0] == "[" and second_value[-1] == "]":
+                second_value = self.__splitList(second_value)
+            processed_datas = self.__processRrd(first_value, second_value, evaluated_dict, sheet_dict)
+            processed_datas = data_type(processed_datas)
+
+        elif prefix == "Renv":
+            processed_datas = self.get_env_variable(raw_data)
+
         elif "-" in raw_data:
             raw_data = raw_data.split('-')
             start = raw_data[0].strip()
@@ -419,9 +500,17 @@ class TestDataGenerator:
                     raw_string = self.__process_rrd_string(raw_string)
                     raw_string = raw_string[4:]
                     data_type = tuple
+                elif raw_string[:4].lower() == "rre_":         # Remote Random (Remote = other sheet)
+                    prefix = "Rre"
+                    raw_string = self.__process_rre_string(raw_string)
+                    raw_string = raw_string[4:]
+                    data_type = tuple
                 else:
                     data_type = list
             else:
+                if raw_string[:5].lower() == "renv_":
+                    prefix = "Renv"
+                    raw_string = raw_string[5:]
                 data_type = list
         else:
             data_type = list
@@ -469,7 +558,7 @@ class TestDataGenerator:
         proccesed_datas = [data.strip() for data in raw_data[1:-1].split(",")]
         return proccesed_datas
 
-    def __processRrd(self, sheet_name, data_looking_for, data_to_match: dict):
+    def __processRrd(self, sheet_name, data_looking_for, data_to_match: dict, sheet_dict=None, caller="RRD_"):
         """
         This function is internal function to process the data wil RRD_ prefix.
         The General input in excel file is like ``RRD_[sheetName,TargetData,[Header:[values**],Header:[values**]]]``
@@ -493,27 +582,70 @@ class TestDataGenerator:
         :param data_to_match:
         :return: dictionary of TargetData
         """
+        sheet_dict = self.sheet_dict if sheet_dict is None else sheet_dict
         matching_data = [list(x) for x in itertools.product(*[data_to_match[key] for key in data_to_match])]
-        assert sheet_name in self.sheet_dict, f"Excel file doesn't contain {sheet_name} sheet. Please recheck. Called in 'RRD_'"
-        base_sheet = self.sheet_dict[sheet_name]
+        assert sheet_name in sheet_dict, \
+            f"Excel file doesn't contain {sheet_name} sheet. Please recheck. Called in '{caller}'"
+        base_sheet = sheet_dict[sheet_name]
         data_lis = []
         if type(data_looking_for) == str:
             data_looking_for = data_looking_for.split(",")
 
+        usecount, limit, usecount_header = self.check_usecount(base_sheet[0])
         for data in base_sheet:
+            if usecount_header:
+                used_limit = data[usecount_header]
+            else:
+                used_limit = 0
             if len(matching_data) == 1 and len(matching_data[0]) == 0:
                 if data_looking_for[0] == "*":
                     data_lis.append(data)
+                    self.usecount_dict[repr(data)] = {
+                        "use": used_limit, "limit": limit, "index": base_sheet.index(data) + 2, "sheet_name": sheet_name
+                    }
                 else:
-                    data_lis.append({keys: data[keys] for keys in data_looking_for})
+                    dt = {keys: data[keys] for keys in data_looking_for}
+                    data_lis.append(dt)
+                    self.usecount_dict[repr(dt)] = {
+                        "use" : used_limit, "limit" : limit, "index": base_sheet.index(data) + 2, "sheet_name": sheet_name
+                    }
             else:
                 if [data[key] for key in data_to_match] in matching_data:
                     if data_looking_for[0] == "*":
                         data_lis.append(data)
+                        self.usecount_dict[repr(data)] = {
+                        "use": used_limit, "limit": limit, "index": base_sheet.index(data) + 2, "sheet_name": sheet_name
+                    }
                     else:
-                        data_lis.append({keys: data[keys] for keys in data_looking_for})
+                        dt = {keys: data[keys] for keys in data_looking_for}
+                        data_lis.append(dt)
+                        self.usecount_dict[repr(dt)] = {
+                        "use" : used_limit, "limit" : limit, "index": base_sheet.index(data) + 2, "sheet_name": sheet_name
+                    }
         return data_lis
 
+    def check_usecount(self, data):
+        # used to find and return if their is usecount header and limit in input file
+        usecount = False
+        limit = 0
+        usecount_header = None
+        for header in data:
+            if "usecount" in header.lower():
+                usecount = True
+                usecount_header = header
+                if "usecount_" in header.lower():
+                    try:
+                        limit = int(header.lower().strip().split("count_")[1])
+                    except:
+                        limit = 0
+        return usecount, limit, usecount_header
+
+    def update_usecount_in_source(self, data):
+        self.writer.write(
+            self.usecount_dict[repr(data)]["index"], self.usecount_dict[repr(data)]["use"],
+            self.usecount_dict[repr(data)]["sheet_name"]
+        )
+
     def __process_rrd_string(self, rrd_string):
         """
         This method is used to validate rrd_strings provided by the user.
@@ -539,6 +671,44 @@ class TestDataGenerator:
         assert match, err_string
         return processed_string
 
+    def __process_rre_string(self, rrd_string):
+        """
+        This method is used to validate rrd_strings provided by the user.
+        If their will be any error in string this fuction will immediately create an error and will stop further execution.
+        Also these function will remove empty spaces around the commas in string.
+        Regex supporting formats in this method are:
+        ``RRE_[fileName,sheetName,TargetData,[Header:[values**],Header:[values**]]]``
+        ``RRE_[fileName,sheetName,[TargetData**],[Header:[values**],Header:[values**]]]``
+        ``RRE_(fileName,sheetName,[TargetData**],[Header:[values**],Header:[values**]])``
+        ``RRE_[fileName,sheetName,*,[Header:[values**],Header:[values**]]]``
+        ``RRE_[fileName,sheetName,*,[Header:[values**],Header:[values**]]]``
+        ``RRE_[fileName,sheetName,TargetData,[]]``
+        ``RRE_(fileName,sheetName,TargetData,[])``
+        ``RRE_(fileName,sheetName,*,[])``
+        ``RRE_[fileName,sheetName,*,[]]``
+        :param rrd_string:
+        :return:
+        """
+        processed_string = ','.join([word.strip() for word in rrd_string.split(', ')])
+        match = re.match(r"(RRE_(\(|\[))[\w\d\s\-./\\]+\.(xlsx|xls),[a-zA-z0-9\s]+,(\[?[a-zA-z\s,]+\]?|)|\*,\[([a-zA-z0-9\s]+:\[[a-zA-z0-9,\s]+\](,?))*\]",processed_string)
+        err_string = f"{rrd_string} not matching pattern RRE_(fileName, sheetName, TargetData," \
+                     f"[Header1:[Value1],Header2:[Value1,Value2]])"
+        assert match, err_string
+        return processed_string
+
+    @staticmethod
+    def get_env_variable(string):
+        variable = string[1:-1].strip().split(',')[0].strip()
+        data = os.environ.get(variable)
+        try:
+            if not data:
+                data = string[1:-1].strip().split(',')[1].strip()
+                logger.info(f"{variable} not found in environment, using {data} instead")
+        except:
+            raise BaseException(f"Can't find {variable} in envrionment & default value is also not set")
+        return data
+
+
 if __name__ == "__main__":
     lTestDataGenerator = TestDataGenerator("../../tests/0TestInput/RawTestData.xlsx")
     lTestDataGenerator.write()
