@@ -7,6 +7,9 @@ from PyQt5 import QtWidgets
 # from baangt.ui.pyqt.uimain import MainWindow
 from baangt.ui.pyqt.uimain import MainController
 from baangt.base.RuntimeStatistics import Statistic
+from baangt.reports import Dashboard, Summary
+import platform
+
 def print_args():
     print("""
 Call: python baangtIA.py --parameters 
@@ -30,11 +33,15 @@ def args_read(l_search_parameter):
         opts, args = getopt.getopt(l_args, "", ["run=",
                                                 "globals=",
                                                 "reloadDrivers=",
-                                                "gui="
+                                                "gui=",
+                                                "cleanup=",
+                                                "name=",
+                                                "stage=",
+                                                "id="
                                                 ])
     except getopt.GetoptError as err_det:
         print("Error in reading parameters:" + str(err_det))
-        print_args()
+        #print_args()
         sys.exit("Wrong parameters - exiting")
     if opts:
         for opt, arg in opts:
@@ -78,8 +85,39 @@ def run():
         lDriver.downloadDriver(GC.BROWSER_FIREFOX)
         lDriver.downloadDriver(GC.BROWSER_CHROME)
         print("Latest versions of drivers for Firefox and Chrome were downloaded")
+
+    # Reports
+    elif args_read("name") or args_read("stage"):
+        name = args_read("name")
+        name = None if name == 'all' else name 
+        
+        try:
+            r = Dashboard(name=name, stage=args_read("stage"))
+            r.show()
+        except ValueError as e:
+            print(f'ERROR: {e}')
+            sys.exit('Exiting...')
+    
+    elif args_read("id"):
+        try:
+            r = Summary(args_read("id"))
+            r.show()
+        except ValueError as e:
+            print(f'ERROR: {e}')
+            sys.exit('Exiting...')
+        
+
+    elif args_read("cleanup"):
+        from baangt.base.Cleanup import Cleanup
+        days = float(args_read("cleanup"))
+        clean = Cleanup(days)
+        clean.clean_all()
     else:
         app = QtWidgets.QApplication(sys.argv)
+        if platform.system() == "Linux":
+            QtWidgets.QApplication.setStyle(QtWidgets.QStyleFactory.create('Fusion'))
+        elif platform.system() == "Darwin":
+            QtWidgets.QApplication.setStyle(QtWidgets.QStyleFactory.create('Windows'))
         controller = MainController()
         controller.show_main()
         sys.exit(app.exec_())
