@@ -4,6 +4,7 @@ import json
 import baangt.base.GlobalConstants as GC
 from baangt.base.Timing.Timing import Timing
 from baangt.base.Utils import utils
+from baangt.base.ExportResults.Append2BaseXLS import Append2BaseXLS
 from pathlib import Path
 from typing import Optional
 from xlsxwriter.worksheet import (
@@ -20,6 +21,9 @@ import csv
 from dateutil.parser import parse
 from uuid import uuid4
 from pathlib import Path
+from baangt.base.ExportResults.SendStatistics import Statistics
+from baangt.base.RuntimeStatistics import Statistic
+from openpyxl import load_workbook
 
 logger = logging.getLogger("pyC")
 
@@ -28,6 +32,7 @@ class ExportResults:
     def __init__(self, **kwargs):
         self.kwargs = kwargs
         self.testList = []
+        self.fileName = None
         self.testRunInstance = kwargs.get(GC.KWARGS_TESTRUNINSTANCE)
         self.testCasesEndDateTimes_1D = kwargs.get('testCasesEndDateTimes_1D')
         self.testCasesEndDateTimes_2D = kwargs.get('testCasesEndDateTimes_2D')
@@ -35,6 +40,10 @@ class ExportResults:
         self.testRunName = self.testRunInstance.testRunName
         self.dataRecords = self.testRunInstance.dataRecords
         self.stage = self.__getStageFromDataRecordsOrGlobalSettings()
+        self.statistics = Statistics()
+        self.statistics.update_data(kwargs)
+        self.statistics.update_runtimeStatistic()
+        self.logfile = logger.handlers[1].baseFilename
 
         try:
             self.exportFormat = kwargs.get(GC.KWARGS_TESTRUNATTRIBUTES).get(GC.EXPORT_FORMAT)
@@ -46,9 +55,22 @@ class ExportResults:
         except KeyError:
             self.exportFormat = GC.EXP_XLSX
 
-        self.fileName = self.__getOutputFileName()
+        try:
+            if kwargs.get(GC.KWARGS_TESTRUNATTRIBUTES).get(GC.STRUCTURE_TESTCASESEQUENCE)[1][1].get(GC.EXPORT_FILENAME):
+                self.fileName = kwargs.get(GC.KWARGS_TESTRUNATTRIBUTES).get(GC.STRUCTURE_TESTCASESEQUENCE)[1][1].get(GC.EXPORT_FILENAME)
+        except Exception as e:
+            # fixme: I don't know, why this error came. When a Filename is set, then the above works.
+            #        No time now to debug.
+            pass
+
+        if not self.fileName:
+            self.fileName = self.__getOutputFileName()
+
         logger.info("Export-Sheet for results: " + self.fileName)
 
+        self.__removeUnwantedFields()  # Will remove Password-Contents AND fields from data records, that came from
+                                       # Globals-File.
+
         # export results to DB
         self.testcase_uuids = []
         self.exportToDataBase()
@@ -72,6 +94,7 @@ class ExportResults:
             self.exportResultExcel()
             self.exportJsonExcel()
             self.exportAdditionalData()
+            self.write_json_sheet()
             self.exportTiming = ExportTiming(self.dataRecords,
                                              self.timingSheet)
             if self.networkInfo:
@@ -82,9 +105,47 @@ class ExportResults:
                                                    self.workbook,
                                                    self.networkSheet)
             self.closeExcel()
+
+            # Call functionality for potentially exporting data to other sheets/databases
+            Append2BaseXLS(self.testRunInstance, self.fileName)
+
         elif self.exportFormat == GC.EXP_CSV:
             self.export2CSV()
-        #self.exportToDataBase()
+
+        if self.testRunInstance.globalSettings.get("DeactivateStatistics") == "True":
+            logger.debug("Send Statistics to server is deactivated. Not sending.")
+        elif self.testRunInstance.globalSettings.get("DeactivateStatistics") is True:
+            logger.debug("Send Statistics to server is deactivated. Not sending.")
+        else:
+            try:
+                self.statistics.send_statistics()
+            except Exception as ex:
+                logger.debug(ex)
+        self.update_result_in_testrun()
+
+    def __removeUnwantedFields(self):
+        lListPasswordFieldNames = ["PASSWORD", "PASSWORT", "PASSW"]
+        if not self.testRunInstance.globalSettings.get("LetPasswords"):
+            # If there's a password in GlobalSettings, remove the value:
+            for key, value in self.testRunInstance.globalSettings.items():
+                if key.upper() in lListPasswordFieldNames:
+                    self.testRunInstance.globalSettings[key] = "*" * 8
+
+            # If there's a password in the datafile, remove the value
+            # Also remove all columns, that are anyway included in the global settings
+            for key, fields in self.dataRecords.items():
+                fieldsToPop = []
+                for field, value in fields.items():
+                    if field.upper() in lListPasswordFieldNames:
+                        self.dataRecords[key][field] = "*" * 8
+                    if field in self.testRunInstance.globalSettings.keys():
+                        logger.debug(
+                            f"Added {field} to fields to be removed from data record as it exists in GlobalSettings already.")
+                        fieldsToPop.append(field)
+                for field in fieldsToPop:
+                    if field != 'Screenshots' and field != 'Stage':   # Stage and Screenshot are needed in output file
+                        logger.debug(f"Removed field {field} from data record.")
+                        fields.pop(field)
 
     def exportAdditionalData(self):
         # Runs only, when KWARGS-Parameter is set.
@@ -112,29 +173,7 @@ class ExportResults:
             stage = value.get(GC.EXECUTION_STAGE)
 
         return stage
-
-    # -- API support --
-    def getSummary(self):
-        #
-        # returns records status as dict
-        #
-        summary = {'Testrecords': len(self.dataRecords)}
-        summary['Successful'] = len([x for x in self.dataRecords.values()
-                                     if x[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_SUCCESS])
-        summary['Paused'] = len([x for x in self.dataRecords.values()
-                                 if x[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_WAITING])
-        summary['Error'] = len([x for x in self.dataRecords.values()
-                                if x[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR])
-        # logfile
-        summary['Logfile'] = logger.handlers[1].baseFilename
-        # timing
-        timing: Timing = self.testRunInstance.timing
-        summary['Starttime'], summary['Endtime'], summary['Duration'] = timing.returnTimeSegment(GC.TIMING_TESTRUN)
-        summary['Globals'] = {key: value for key, value in self.testRunInstance.globalSettings.items()}
-
-        return summary
-
-    # -- END of API support --
+        
 
     def export2CSV(self):
         """
@@ -174,7 +213,19 @@ class ExportResults:
                 error += 1
             if value[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_WAITING:
                 waiting += 1
-
+        self.statistics.update_attribute_with_value("TestCasePassed", success)
+        self.statistics.update_attribute_with_value("TestCaseFailed", error)
+        self.statistics.update_attribute_with_value("TestCasePaused", waiting)
+        self.statistics.update_attribute_with_value("TestCaseExecuted", success + error + waiting)
+        try:
+            dic = {}
+            for key in self.testRunInstance.json_dict:
+                if "$(" in key:
+                    dic[key[2:-1]] = self.testRunInstance.json_dict[key]
+            json_data = json.dumps(dic)
+        except Exception as ex:
+            logger.info(f"RLP Json error while updating in db : {str(ex)}")
+            json_data = ""
         # get documents
         datafiles = self.fileName
 
@@ -182,13 +233,14 @@ class ExportResults:
         tr_log = TestrunLog(
             id=self.testRunInstance.uuid.bytes,
             testrunName=self.testRunName,
-            logfileName=logger.handlers[1].baseFilename,
+            logfileName=self.logfile,
             startTime=datetime.strptime(start, "%d-%m-%Y %H:%M:%S"),
             endTime=datetime.strptime(end, "%d-%m-%Y %H:%M:%S"),
             statusOk=success,
             statusFailed=error,
             statusPaused=waiting,
-            dataFile=datafiles,
+            dataFile=self.fileName,
+            RLPJson=json_data,
         )
         # add to DataBase
         session.add(tr_log)
@@ -332,6 +384,55 @@ class ExportResults:
         for n in range(len(headers)):
             ExcelSheetHelperFunctions.set_column_autowidth(self.jsonSheet, n)
 
+    def write_json_sheet(self):
+        # Used to write rlp_ json in individual sheets
+        dic = self.testRunInstance.json_dict
+        for js in dic:
+            if not js:
+                continue
+            elif js[:2] == "$(":
+                name = js[2:-1]
+            else:
+                name = js
+            jsonSheet = self.workbook.add_worksheet(f"{self.stage}_{name}")
+            if type(dic[js][0]) == dict: # Condition to get dictionary or dictionary inside list to write headers
+                data_dic = dic[js][0]
+            elif type(dic[js][0][0]) == dict:
+                data_dic = dic[js][0][0]
+            else:
+                logger.debug(f"{dic[js]} is not json convertible.")
+                continue
+            remove_header = []
+            for key in data_dic: # Removing headers which consist nested data
+                if type(data_dic[key]) == list or type(data_dic[key]) == dict:
+                    remove_header.append(key)
+            for key in remove_header:
+                del data_dic[key]
+            headers = []
+            for index, header in enumerate(data_dic):
+                jsonSheet.write(0, index, header)
+                headers.append(header)
+            row = 1
+            for data in dic[js]:
+                if not data:
+                    continue
+                dt = {}
+                for y, dt in enumerate(data):
+                    if type(dt) != dict: # for single dictionary data
+                        jsonSheet.write(row, y, data[dt])
+                    else: # if dictionaries are inside list
+                        column = 0 # used to update individual column
+                        for d in dt:
+                            if d not in headers:
+                                continue
+                            try:
+                                jsonSheet.write(row, column, dt[d])
+                            except Exception as ex:
+                                print(ex)
+                            column += 1
+                        row += 1
+                if type(dt) != dict:
+                    row += 1
 
     def makeSummaryExcel(self):
 
@@ -366,6 +467,9 @@ class ExportResults:
         # Timing
         timing: Timing = self.testRunInstance.timing
         start, end, duration = timing.returnTimeSegment(GC.TIMING_TESTRUN)
+        self.testRun_end = end  # used while updating timestamp in source file
+        self.statistics.update_attribute_with_value("Duration", duration)
+        self.statistics.update_attribute_with_value("TestRunUUID", str(self.testRunInstance.uuid))
         self.__writeSummaryCell("Starttime", start, row=10)
         # get start end during time my
         self.testList.append(start)
@@ -377,6 +481,8 @@ class ExportResults:
         # Globals:
         self.__writeSummaryCell("Global settings for this testrun", "", format=self.cellFormatBold, row=15)
         for key, value in self.testRunInstance.globalSettings.items():
+            if key.upper() in ["PASSWORD", "PASSWORT", "CONFLUENCE-PASSWORD"]:
+                continue
             self.__writeSummaryCell(key, str(value))
             # get global data my
             self.testList.append(str(value))
@@ -487,6 +593,25 @@ class ExportResults:
         for n in range(0, len(self.fieldListExport)):
             ExcelSheetHelperFunctions.set_column_autowidth(self.worksheet, n)
 
+    def update_result_in_testrun(self):
+        # To update source testrun file
+        testrun_column = self.dataRecords[0]["testcase_column"]
+        if testrun_column:  # if testrun_column is greater than 0 that means testresult header is present in source file
+            testrun_file = load_workbook(self.dataRecords[0]["testcase_file"])
+            testrun_sheet = testrun_file.get_sheet_by_name(self.dataRecords[0]["testcase_sheet"])
+            for key, value in self.dataRecords.items():
+                print(value)
+                data = f"TestCaseStatus: {value['TestCaseStatus']}\r\n" \
+                       f"Timestamp: {self.testRun_end}\r\n" \
+                       f"Duration: {value['Duration']}\r\n" \
+                       f"TCErrorLog: {value['TCErrorLog']}\r\n" \
+                       f"TestRun_UUID: {str(self.testRunInstance.uuid)}\r\n" \
+                       f"TestCase_UUID: {str(self.testcase_uuids[key])}\r\n\r\n"
+                old_value = testrun_sheet.cell(value["testcase_row"] + 1, value["testcase_column"]).value or ""
+                testrun_sheet.cell(value["testcase_row"] + 1, value["testcase_column"]).value = data + old_value
+            testrun_file.save(self.dataRecords[0]["testcase_file"])
+            logger.info(f"Source TestRun file {self.dataRecords[0]['testcase_file']} updated.")
+
     def __writeCell(self, line, cellNumber, testRecordDict, fieldName, strip=False):
         if fieldName in testRecordDict.keys() and testRecordDict[fieldName]:
             # Convert boolean for Output
@@ -494,8 +619,10 @@ class ExportResults:
                 testRecordDict[fieldName] = "True" if testRecordDict[fieldName] else "False"
 
             # Remove leading New-Line:
-            if '\n' in testRecordDict[fieldName][0:5] or strip:
-                testRecordDict[fieldName] = testRecordDict[fieldName].strip()
+            if isinstance(testRecordDict[fieldName], str):
+                if '\n' in testRecordDict[fieldName][0:5] or strip:
+                    testRecordDict[fieldName] = testRecordDict[fieldName].strip()
+
             # Do different stuff for Dicts and Lists:
             if isinstance(testRecordDict[fieldName], dict):
                 self.worksheet.write(line, cellNumber, testRecordDict[fieldName])
@@ -508,34 +635,34 @@ class ExportResults:
             else:
                 if fieldName == GC.TESTCASESTATUS:
                     if testRecordDict[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_SUCCESS:
-                        self.worksheet.write(line, cellNumber, testRecordDict[fieldName], self.cellFormatGreen)
+                        self.worksheet.write(line, cellNumber, str(testRecordDict[fieldName]), self.cellFormatGreen)
                     elif testRecordDict[GC.TESTCASESTATUS] == GC.TESTCASESTATUS_ERROR:
-                        self.worksheet.write(line, cellNumber, testRecordDict[fieldName], self.cellFormatRed)
+                        self.worksheet.write(line, cellNumber, str(testRecordDict[fieldName]), self.cellFormatRed)
                 elif fieldName == GC.SCREENSHOTS:
                     self.__attachScreenshotsToExcelCells(cellNumber, fieldName, line, testRecordDict)
                 else:
-                    self.worksheet.write(line, cellNumber, testRecordDict[fieldName])
+                    self.worksheet.write(line, cellNumber, str(testRecordDict[fieldName]))
 
     def __attachScreenshotsToExcelCells(self, cellNumber, fieldName, line, testRecordDict):
         # Place the screenshot images "on" the appropriate cell
         try:
             if type(testRecordDict[fieldName]) == list:
 
-                if Path(testRecordDict[fieldName][-1]).exists():
+                if Path(testRecordDict[fieldName][-1]).is_file():
                     self.worksheet.insert_image(line, cellNumber, testRecordDict[fieldName][-1], {'x_scale': 0.05,
                                                                                                   'y_scale': 0.05})
                 else:
                     logger.error(f"Sceenshot file {testRecordDict[fieldName][-1]} can't be found")
 
                 for nextScreenshotNumber in range(len(testRecordDict[fieldName]) - 1):
-                    if Path(testRecordDict[fieldName][nextScreenshotNumber]).exists():
+                    if Path(testRecordDict[fieldName][nextScreenshotNumber]).is_file():
                         self.worksheet.insert_image(line, len(self.fieldListExport) + nextScreenshotNumber + 1,
                                                     testRecordDict[fieldName][nextScreenshotNumber],
                                                     {'x_scale': 0.05, 'y_scale': 0.05})
                     else:
                         logger.error(f"Screenshot file {testRecordDict[fieldName][nextScreenshotNumber]} can't be found")
             else:
-                if Path(testRecordDict[fieldName]).exists():
+                if Path(testRecordDict[fieldName]).is_file():
                     self.worksheet.insert_image(line, cellNumber, testRecordDict[fieldName], {'x_scale': 0.05,
                                                                                               'y_scale': 0.05})
                 else:
