@@ -27,9 +27,14 @@ from baangt.base.FilesOpen import FilesOpen
 from baangt.base.RuntimeStatistics import Statistic
 from baangt.base.PathManagement import ManagedPaths
 from baangt.base.DownloadFolderMonitoring import DownloadFolderMonitoring
-import xlrd
+from baangt.base.Cleanup import Cleanup
+import xlrd3 as xlrd
+from baangt.reports import Dashboard, Summary
+from baangt.TestDataGenerator.TestDataGenerator import TestDataGenerator
+from baangt.base.Utils import utils
 from threading import Thread
 from time import sleep
+import signal
 
 logger = logging.getLogger("pyC")
 
@@ -83,6 +88,9 @@ class MainWindow(Ui_MainWindow):
         self.__execute_button_state = "idle"
         self.__result_file = ""
         self.__log_file = ""
+        self.__open_files = 0
+        self.TDGResult = ""
+        self.dataFile = ""
 
         # self.refreshNew()
         # self.setupBasePath(self.directory)
@@ -98,6 +106,12 @@ class MainWindow(Ui_MainWindow):
         # initialize Katalon Importer and Global Setting Page
         # Add Button Signals and Slot here
         self.browsePushButton_4.clicked.connect(self.browsePathSlot)
+        self.InputFileButton.clicked.connect(self.inputFileSlot)
+        self.OutputFileButton.clicked.connect(self.outputFileSlot)
+        self.exitPushButton_4.clicked.connect(self.mainPageView)
+        self.TDGButton.clicked.connect(self.testDataGenerator)
+        self.ResultButton.clicked.connect(self.openTDGResult)
+        self.SheetCombo.activated.connect(self.updateSheet)
         # self.executePushButton.clicked(self.executeTest)
 
         # Setting Page actions and triggered
@@ -115,16 +129,28 @@ class MainWindow(Ui_MainWindow):
         self.openResultFilePushButton_4.clicked.connect(self.openResultFile)
         self.openLogFilePushButton_4.clicked.connect(self.openLogFile)
         self.openTestFilePushButton_4.clicked.connect(self.openTestFile)
+        self.InputFileOpen.clicked.connect(self.openInputFile)
 
         # Quit Event
         self.actionExit.triggered.connect(self.quitApplication)
 
+        # Show Report Event
+        self.actionReport.triggered.connect(self.showReport)
+
+        # Cleanup action
+        self.actionCleanup.triggered.connect(self.cleanup_dialog)
+
+        # TestDataGenerator
+        self.actionTestDataGen.triggered.connect(self.show_tdg)
+
         # Katalon triggered
         self.actionImport_Katalon.triggered.connect(self.show_katalon)
         self.exitPushButton_3.clicked.connect(self.exitKatalon)
         self.savePushButton_2.clicked.connect(self.saveTestCase)
         self.copyClipboard_2.clicked.connect(self.copyFromClipboard)
         self.TextIn_2.textChanged.connect(self.importClipboard)
+
+        # log & open file switch
         self.logSwitch.clicked.connect(self.show_hide_logs)
         self.openFilesSwitch.clicked.connect(self.change_openFiles_state)
 
@@ -202,6 +228,12 @@ class MainWindow(Ui_MainWindow):
         self.stackedWidget.setCurrentIndex(2)
         self.statusMessage("Katalon Studio is triggered", 1000)
 
+    @QtCore.pyqtSlot()
+    def show_tdg(self):
+        """ Display katalon panel for Test case preparation """
+        self.stackedWidget.setCurrentIndex(3)
+        self.statusMessage("TestDataGenerator is triggered", 1000)
+
     def updateLogoAndIcon(self, MainWindow):
         """ This function initialize logo and icon """
         logo_pixmap = QtGui.QPixmap(":/baangt/baangtlogo")
@@ -348,6 +380,46 @@ class MainWindow(Ui_MainWindow):
         self.getSettingsAndTestFilesInDirectory(dirPath)
         self.statusMessage("Current Path: {} ".format(dirPath), 2000)
 
+    def setupFilePath(self, filePath=()):
+        """ Setup Base path of Execution as per directory Path"""
+
+        if os.path.exists(filePath[0]):
+            dirPath = filePath[0]
+            self.InputFileEdit.setText(dirPath)
+            self.directory = dirPath
+            self.getSheets(dirPath)
+            self.OutputFileEdit.setText(os.path.dirname(dirPath))
+            self.statusMessage("Current File: {} ".format(dirPath), 2000)
+
+    def getSheets(self, dirName):
+        """ Scan for *.xlsx files and *.json files and
+        update the testRunComboBox and settingsComboBox items
+        """
+        wb = xlrd.open_workbook(dirName)
+        self.Sheets = wb.sheet_names()
+        self.SheetCombo.clear()
+        # Add files in Combo Box
+        self.SheetCombo.addItems(self.Sheets)
+        # set default selection to 0
+
+        if len(self.Sheets) > 0:
+            try:
+                index = self.testRunComboBox_4.findText(
+                                     self.selectedSheet,
+                                     QtCore.Qt.MatchFixedString
+                                     )
+            except:
+                self.selectedSheet = ""
+                index = 0
+            if self.selectedSheet not in self.Sheets:
+                self.SheetCombo.setCurrentIndex(0)
+                self.selectedSheet = self.SheetCombo.currentText()
+            else:
+                self.SheetCombo.setCurrentIndex(index)
+            self.statusMessage("Sheet: {}".format(self.selectedSheet), 3000)
+            # Activate the Execute Button
+            self.TDGButton.setEnabled(True)
+
     @pyqtSlot()
     def quitApplication(self):
         """ This function will close the UI """
@@ -371,9 +443,8 @@ class MainWindow(Ui_MainWindow):
     def updateRunFile(self):
         """ this file will update the testRunFile selection
         """
-        self.testRunFile = os.path.join(self.directory,
-                                  self.testRunComboBox_4.currentText())
-        self.statusMessage("Test Run Changed to: {}".format(self.testRunFile))
+        self.testRunFile = self.testRunComboBox_4.currentText()
+        self.statusMessage("Test Run Changed to: {}".format(self.testRunFile), 3000)
         self.saveInteractiveGuiConfig()
 
     @pyqtSlot()
@@ -385,9 +456,16 @@ class MainWindow(Ui_MainWindow):
                                  self.settingComboBox_4.currentText())
 
         self.saveInteractiveGuiConfig()
-        self.statusMessage("Settings changed to: {}".format(self.configFile))
+        self.statusMessage("Settings changed to: {}".format(self.configFile), 3000)
         self.readContentofGlobals()
 
+    @pyqtSlot()
+    def updateSheet(self):
+        """ this file will update the testRunFile selection
+        """
+        self.selectedSheet = self.SheetCombo.currentText()
+        self.statusMessage("Selected Sheet: {}".format(self.selectedSheet), 3000)
+
     def executeButtonClicked(self):
         self.__result_file = ""
         self.__log_file = ""
@@ -410,7 +488,7 @@ class MainWindow(Ui_MainWindow):
         self.stopIcon = QtGui.QIcon(":/baangt/stopicon")
         self.executePushButton_4.setIcon(self.stopIcon)
         self.executePushButton_4.setIconSize(QtCore.QSize(28, 20))
-        self.executePushButton_4.setStyleSheet("color: rgb(255, 255, 255); background-color: rgb(255, 75, 50);")
+        self.executePushButton_4.setStyleSheet("color: rgb(255, 255, 255); background-color: rgb(204, 0, 0);")
         self.clear_logs_and_stats()
 
         runCmd = self._getRunCommand()
@@ -438,11 +516,108 @@ class MainWindow(Ui_MainWindow):
                     str(self.run_process.readAllStandardError().data().decode('iso-8859-1'))))
             self.run_process.finished.connect(self.processFinished)
             self.run_process.start(runCmd)
-            self.statusbar.showMessage("Running.....")
+            self.statusbar.showMessage("Running.....",4000)
 
     @pyqtSlot()
     def stopButtonPressed(self):
-        self.run_process.kill()
+        if platform.system() == "Windows":
+            self.signalCtrl(self.run_process)
+            try:
+                os.remove(Path(self.directory).joinpath(self.tempConfigFile))
+            except Exception as e:
+                logger.warning(f"Tried to remove temporary file but seems to be not there: "
+                               f"{self.directory}/{self.tempConfigFile}")
+            self.run_process.kill()
+            QtWidgets.QApplication.quit()
+        else:
+            os.kill(self.run_process.processId(), signal.SIGINT)
+            self.run_process.waitForFinished(3000)
+            self.run_process.kill()
+
+    def update_datafile(self):
+        from baangt.base.TestRun.TestRun import TestRun
+        testRunFile = f"{Path(self.directory).joinpath(self.testRunFile)}"
+        globalsFile = self.configFile
+        uu = uuid4()
+        tr = TestRun(testRunFile, globalsFile, uuid=uu, executeDirect=False)
+        tr._initTestRunSettingsFromFile()
+        if "TC.TestDataFileName" in tr.globalSettings:
+            self.dataFile = tr.globalSettings["TC.TestDataFileName"]
+        else:
+            tr._loadJSONTestRunDefinitions()
+            tr._loadExcelTestRunDefinitions()
+            self.dataFile = self.findKeyFromDict(tr.testRunUtils.testRunAttributes, "TestDataFileName")
+
+    def findKeyFromDict(self, dic, key):
+        if isinstance(dic, list):
+            for data in dic:
+                if isinstance(dic, list) or isinstance(dic, dict):
+                    result = self.findKeyFromDict(data, key)
+                    if result:
+                        return result
+        elif isinstance(dic, dict):
+            for k in dic:
+                if k == key:
+                    return dic[k]
+                elif isinstance(dic[k], list) or isinstance(dic[k], dict):
+                    result = self.findKeyFromDict(dic[k], key)
+                    if result:
+                        return result
+        return ""
+
+
+
+    def signalCtrl(self, qProcess, ctrlEvent=None):
+        import win32console, win32process, win32api, win32con
+        if ctrlEvent is None:
+            ctrlEvent = win32con.CTRL_C_EVENT
+
+        has_console = False
+        try:
+            win32console.AllocConsole()
+        except win32api.error:
+            has_console = True
+        if not has_console:
+            # free the dummy console
+            try:
+                win32console.FreeConsole()
+            except win32api.error:
+                return False
+
+        if has_console:
+            # preserve the console in a dummy process
+            try:
+                hProc, _, pid, _ = win32process.CreateProcess(
+                    None, "cmd.exe", None, None, True, win32con.DETACHED_PROCESS,
+                    None, 'c:\\', win32process.STARTUPINFO())
+                win32console.FreeConsole()
+            except win32api.error:
+                return False
+
+        try:
+            # attach to the process's console and generate the event
+            win32console.AttachConsole(qProcess.processId())
+            # Add a fake Ctrl-C handler for avoid instant kill is this console
+            win32api.SetConsoleCtrlHandler(None, True)
+            win32console.GenerateConsoleCtrlEvent(ctrlEvent, 0)
+            sleep(1)
+            win32console.GenerateConsoleCtrlEvent(ctrlEvent, 0)
+            win32console.FreeConsole()
+        except win32api.error:
+            return False
+
+        if not has_console:
+            # we have no console to restore
+            return True
+
+        try:
+            # attach to the dummy process's console and end the process
+            win32console.AttachConsole(pid)
+            win32process.TerminateProcess(hProc, 1)
+        except Exception as ex:
+            return False
+        return True
+
 
     @pyqtSlot()
     def processFinished(self, debug=False):
@@ -487,7 +662,8 @@ class MainWindow(Ui_MainWindow):
                 else:
                     break
             if result_file[-5:] == ".xlsx":
-                self.__result_file =result_file
+                logger.debug(f"Found result_file in the logs: {result_file}")
+                self.__result_file = result_file
         if "||Statistic:" in text:
             lis = text.split('||')
             stat = lis[1][10:]
@@ -501,9 +677,9 @@ class MainWindow(Ui_MainWindow):
                 self.statisticTable.item(0, x).setFlags(
                     self.statisticTable.item(0, x).flags() ^ QtCore.Qt.ItemIsEditable)
                 if x == 3:
-                    self.statisticTable.item(0, 3).setBackground(QtGui.QBrush(QtCore.Qt.green))
+                    self.statisticTable.item(0, 3).setBackground(QtGui.QColor(115, 210, 22))#QBrush(QtCore.Qt.green))
                 elif x == 4:
-                    self.statisticTable.item(0, 4).setBackground(QtGui.QBrush(QtCore.Qt.red))
+                    self.statisticTable.item(0, 4).setBackground(QtGui.QColor(204, 0, 0))#QBrush(QtCore.Qt.red))
                 else:
                     self.statisticTable.item(0, x).setBackground(QtGui.QBrush(QtCore.Qt.white))
             log.replace('\n', '')
@@ -602,6 +778,48 @@ class MainWindow(Ui_MainWindow):
             # self.pathLineEdit.insert(dirName)
             self.setupBasePath(dirName)
 
+    @pyqtSlot()
+    def inputFileSlot(self):
+        """ Browse Folder Containing *.xlsx file for execution. And
+           globals.json file for Test specific settings
+        """
+        # get path from pathLineEdit
+        basepath = self.InputFileEdit.text()
+        if not basepath:
+            basepath = "./"
+        options = QtWidgets.QFileDialog.Options()
+        options |= QtWidgets.QFileDialog.DontUseNativeDialog
+        fileName = QtWidgets.QFileDialog.getOpenFileName(
+            None,
+            "Select File ",
+            basepath,
+            "(*.xlsx)",
+            options=options
+        )
+        if fileName:
+            self.setupFilePath(fileName)
+
+    @pyqtSlot()
+    def outputFileSlot(self):
+        """ Browse Folder Containing *.xlsx file for execution. And
+           globals.json file for Test specific settings
+        """
+        # get path from pathLineEdit
+        basepath = self.OutputFileEdit.text()
+        if not basepath:
+            basepath = "./"
+        options = QtWidgets.QFileDialog.Options()
+        options |= QtWidgets.QFileDialog.DontUseNativeDialog
+        dirName = QtWidgets.QFileDialog.getExistingDirectory(
+            None,
+            "Select Directory ",
+            basepath,
+            options=options
+        )
+        if dirName:
+            # self.pathLineEdit.insert(dirName)
+            self.OutputFileEdit.setText(dirName)
+
     # Settings Page
     # All action and function related to Setting page is below
     def readConfigFile(self):
@@ -738,7 +956,7 @@ class MainWindow(Ui_MainWindow):
                 if self.directory:
                     fullpath = os.path.join(self.directory, self.configFile)
                 else:
-                    self.directory = os.getcwd()#.managedPaths.derivePathForOSAndInstallationOption()
+                    self.directory = os.getcwd()   # .managedPaths.derivePathForOSAndInstallationOption()
                     fullpath = os.path.join(self.directory, self.configFile)
 
             with open(fullpath, 'w') as f:
@@ -1021,6 +1239,7 @@ class MainWindow(Ui_MainWindow):
         """ Uses Files Open class to open Result file """
         try:
             if self.__result_file != "":
+                logger.debug(f"Opening ResultFile: {self.__result_file}")
                 filePathName = self.__result_file
                 fileName = os.path.basename(filePathName)
                 self.statusMessage(f"Opening file {fileName}", 3000)
@@ -1058,11 +1277,38 @@ class MainWindow(Ui_MainWindow):
         try:
             filePathName = f"{Path(self.directory).joinpath(self.testRunFile)}"
             fileName = os.path.basename(filePathName)
-            self.statusbar.showMessage(f"Opening file {fileName}")
+            self.statusMessage(f"Opening file {fileName}", 3000)
             FilesOpen.openResultFile(filePathName)
+            self.update_datafile()
+            if self.dataFile:
+                self.statusMessage(f"Opening file {self.dataFile}", 3000)
+                PathName = f"{Path(self.directory).joinpath(self.dataFile)}"
+                Name = os.path.basename(PathName)
+                FilesOpen.openResultFile(PathName)
         except:
             self.statusMessage("No file found!", 3000)
 
+    @QtCore.pyqtSlot()
+    def openInputFile(self):
+        """ Uses Files Open class to open Log file """
+        filePathName = self.InputFileEdit.text()
+        if len(filePathName) > 0:
+            fileName = os.path.basename(filePathName)
+            self.statusMessage(f"Opening file {fileName}", 3000)
+            FilesOpen.openResultFile(filePathName)
+        else:
+            self.statusMessage("Please select a file", 3000)
+
+    @QtCore.pyqtSlot()
+    def openTDGResult(self):
+        """ Uses Files Open class to open Log file """
+        if os.path.exists(self.TDGResult):
+            fileName = os.path.basename(self.TDGResult)
+            self.statusMessage(f"Opening file {fileName}", 3000)
+            FilesOpen.openResultFile(self.TDGResult)
+        else:
+            self.statusMessage("No result file found!", 3000)
+
     @pyqtSlot()
     def clear_logs_and_stats(self):
         for x in range(9):
@@ -1088,6 +1334,77 @@ class MainWindow(Ui_MainWindow):
             self.__open_files = 0
         self.saveInteractiveGuiConfig()
 
+    @pyqtSlot()
+    def cleanup_dialog(self):
+        self.clean_dialog = QtWidgets.QDialog(self.centralwidget)
+        self.clean_dialog.setWindowTitle("Cleanup")
+        vlay = QtWidgets.QVBoxLayout()
+        self.cleanup_logs = QtWidgets.QCheckBox()
+        self.cleanup_logs.setText("Logs")
+        self.cleanup_logs.setChecked(True)
+        self.cleanup_screenshots = QtWidgets.QCheckBox()
+        self.cleanup_screenshots.setText("Screenshots")
+        self.cleanup_screenshots.setChecked(True)
+        self.cleanup_downloads = QtWidgets.QCheckBox()
+        self.cleanup_downloads.setText("Downloads")
+        self.cleanup_downloads.setChecked(True)
+        hlay = QtWidgets.QHBoxLayout()
+        label = QtWidgets.QLabel()
+        label.setText("Days: ")
+        self.cleanup_days = QtWidgets.QLineEdit("31", self.clean_dialog)
+        self.cleanup_days.setValidator(QtGui.QIntValidator())
+        self.cleanup_days.setStyleSheet("background-color: rgb(255, 255, 255);")
+        hlay.addWidget(label)
+        hlay.addWidget(self.cleanup_days)
+        button = QtWidgets.QPushButton("Cleanup", self.clean_dialog)
+        button.setStyleSheet("color: rgb(255, 255, 255); background-color: rgb(204, 0, 0);")
+        self.cleanup_status = QtWidgets.QStatusBar()
+        vlay.addWidget(self.cleanup_logs)
+        vlay.addWidget(self.cleanup_screenshots)
+        vlay.addWidget(self.cleanup_downloads)
+        vlay.addLayout(hlay)
+        vlay.addWidget(button)
+        vlay.addWidget(self.cleanup_status)
+        self.clean_dialog.setLayout(vlay)
+        button.clicked.connect(self.cleanup_button)
+        self.clean_dialog.exec_()
+
+
+    def cleanup_button(self):
+        self.cleanup_status.showMessage("Cleaning...")
+        text = self.cleanup_days.text()
+        if len(text) > 0:
+            days = int(text)
+        else:
+            days = 0
+        c = Cleanup(days)
+        if self.cleanup_logs.isChecked():
+            c.clean_logs()
+        if self.cleanup_screenshots.isChecked():
+            c.clean_screenshots()
+        if self.cleanup_downloads.isChecked():
+            c.clean_downloads()
+        self.cleanup_status.showMessage("Cleaning Complete!")
+
+    def showReport(self):
+        r = Dashboard()
+        r.show()
+
+    def testDataGenerator(self):
+        if len(self.ResultLengthInput.text()) > 0:
+            batch_size = int(self.ResultLengthInput.text())
+        else:
+            batch_size = 0
+        input_file = self.InputFileEdit.text()
+        input_file_name = os.path.basename(input_file)
+        tdg = TestDataGenerator(input_file, self.selectedSheet)
+        outputfile = os.path.join(
+            self.OutputFileEdit.text(),
+            f"{input_file_name}_{self.selectedSheet}_{utils.datetime_return()}.xlsx"
+        )
+        tdg.write(outputfile=outputfile, batch_size=batch_size)
+        self.TDGResult = outputfile
+        self.statusMessage(f"Data Generated Successfully: {outputfile}", 3000)
 
 # Controller
 class MainController:
